/*! Tripetto Collector 0.4.26-26 - Copyright (C) 2018 Tripetto B.V. - All Rights Reserved */

/// <reference path="marblecore.matrix.d.ts" />
/// <reference path="marblecore.callback.d.ts" />
/// <reference path="tripetto.slots.d.ts" />
/// <reference path="marblecore.serializer.d.ts" />
/// <reference path="marblecore.hookup.d.ts" />
/// <reference path="marblecore.functional.d.ts" />

declare module "@tripetto/vfsm" {
    /** Explicit exports (fixed order) */
    export { Conditions } from "@tripetto/vfsm/lib/conditions/conditions";
    export { Condition } from "@tripetto/vfsm/lib/conditions/condition";
    export { Branches } from "@tripetto/vfsm/lib/branches/branches";
    export { Branch } from "@tripetto/vfsm/lib/branches/branch";
    export { States } from "@tripetto/vfsm/lib/states/states";
    export { State } from "@tripetto/vfsm/lib/states/state";
    export { Transducers } from "@tripetto/vfsm/lib/transducers/transducers";
    export { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    /** Interfaces */
    export { IBranch } from "@tripetto/vfsm/lib/branches/interface";
    export { IChainLookup } from "@tripetto/vfsm/lib/chain";
    export { ICondition } from "@tripetto/vfsm/lib/conditions/interface";
    export {
        IDataValue,
        IDataValues,
        IDataImport
    } from "@tripetto/vfsm/lib/instance/data/interface";
    export { IMetrics } from "@tripetto/vfsm/lib/metrics";
    export { ImmutableValue } from "@tripetto/vfsm/lib/instance/data/immutable";
    export { IMoment } from "@tripetto/vfsm/lib/instance/moments/moment";
    export { Instance } from "@tripetto/vfsm/lib/instance/instance";
    export { IOntologyProperties } from "@tripetto/vfsm/lib/interface";
    export { IInstanceSnapshot } from "@tripetto/vfsm/lib/instance/snapshot";
    export { ISnapshot } from "@tripetto/vfsm/lib/snapshot";
    export { IState } from "@tripetto/vfsm/lib/states/interface";
    export { ITransducer } from "@tripetto/vfsm/lib/transducers/interface";
    export { IVariable } from "@tripetto/vfsm/lib/instance/data/variable";
    /** Other explicit exports */
    export { Context } from "@tripetto/vfsm/lib/instance/data/context";
    export { Data } from "@tripetto/vfsm/lib/instance/data/data";
    export { Ontology } from "@tripetto/vfsm/lib/ontology";
    export { Value } from "@tripetto/vfsm/lib/instance/data/value";
    export { Values } from "@tripetto/vfsm/lib/instance/data/values";
    /** Enumerations */
    export { Errors } from "@tripetto/vfsm/enum/errors";
    /** Decorators */
    export { ontology } from "@tripetto/vfsm/lib/decorator";
    /** Events */
    export {
        IOntologyEndEvent,
        IOntologyFinishEvent,
        IOntologyInstanceCreateEvent,
        IOntologyInstanceEndEvent,
        IOntologyInstancePauseEvent,
        IOntologyInstanceStartEvent,
        IOntologyInstanceStopEvent,
        IOntologyLoadEvent,
        IOntologyPauseEvent,
        IOntologyStartEvent,
        IOntologyStopEvent,
        IOntologyUnloadEvent
    } from "@tripetto/vfsm/lib/events";
    export {
        IInstanceCreateEvent,
        IInstanceEndEvent,
        IInstancePauseEvent,
        IInstanceStartEvent,
        IInstanceStopEvent
    } from "@tripetto/vfsm/lib/instance/events";
    export {
        IBranchCancelEvent,
        IBranchEnterEvent,
        IBranchLeaveEvent,
        IBranchReadyEvent,
        IBranchSkipEvent
    } from "@tripetto/vfsm/lib/branches/events";
    export {
        IStateCancelEvent,
        IStateEnterEvent,
        IStateEvaluateEvent,
        IStateLeaveEvent,
        IStateSkipEvent,
        IStateValidatedEvent
    } from "@tripetto/vfsm/lib/states/events";
    export {
        ITransducerCancelEvent,
        ITransducerEnterEvent,
        ITransducerLeaveEvent,
        ITransducerSkipEvent,
        ITransducerValidatedEvent
    } from "@tripetto/vfsm/lib/transducers/events";
    export { IValueChangeEvent } from "@tripetto/vfsm/lib/instance/data/events";
}

declare module "@tripetto/vfsm/lib/conditions/conditions" {
    /** Dependencies */
    import { Branch } from "@tripetto/vfsm/lib/branches/branch";
    import { Condition } from "@tripetto/vfsm/lib/conditions/condition";
    import { ICondition } from "@tripetto/vfsm/lib/conditions/interface";
    import { StrictMatrix } from "@marblecore/matrix";
    export class Conditions extends StrictMatrix<
        Conditions,
        Condition,
        void,
        ICondition
    > {
        /**
         * Constructs a new condition dispatcher.
         * @param Reference to the parent branch.
         */
        constructor(pBranch: Branch);
        /** Retrieves a reference to the parent branch. */
        readonly branch: Branch;
        /**
         * Creates a condition.
         * @param pCondition Reference to the condition ontology.
         * @return Returns a reference to the condition instance or `undefined` if
         * a similar condition was already found.
         */
        create<T extends Condition>(pCondition: ICondition): T | undefined;
    }
}

declare module "@tripetto/vfsm/lib/conditions/condition" {
    /** Dependencies */
    import { Branch } from "@tripetto/vfsm/lib/branches/branch";
    import { Callback } from "@marblecore/callback";
    import { Conditions } from "@tripetto/vfsm/lib/conditions/conditions";
    import { Context } from "@tripetto/vfsm/lib/instance/data/context";
    import { Factory } from "@tripetto/vfsm/lib/factory";
    import { ICondition } from "@tripetto/vfsm/lib/conditions/interface";
    import { State } from "@tripetto/vfsm/lib/states/state";
    export abstract class Condition<
        T extends ICondition = ICondition
    > extends Factory<Conditions, T> {
        /**
         * Retrieves the context for the given condition props.
         * @param pCondition Reference to the condition props.
         * @return Returns the context.
         */
        static getContext<T extends ICondition>(pCondition: T): string;
        /**
         * Constructs a new condition.
         * @param pConditions Reference to the condition dispatcher.
         * @param pCondition Reference to the condition props.
         */
        constructor(pConditions: Conditions, pCondition: T);
        /** Reference to the conditions dispatcher. */
        protected readonly conditions: Conditions;
        /** Reference to the parent branch. */
        readonly branch: Branch;
        /** Reference to the parent state. */
        readonly state: State;
        /** Retrieves the condition context. */
        readonly context: string;
        /** Retrieves the condition properties. */
        readonly props: Readonly<T>;
        /**
         * Invoked when the VFSM wants to verify the condition.
         * @param pContext Reference to the context.
         * @param pCallback Reference to the callback pointer. Pass this pointer as
         * return value to enable the asynchronous condition callback.
         * @return Returns the result of the condition or the callback reference if
         * the condition should be handled asynchronously.
         * @event
         */
        abstract onCondition(
            pContext: Context,
            pCallback: Callback<boolean>
        ): boolean | Callback<boolean>;
    }
}

declare module "@tripetto/vfsm/lib/branches/branches" {
    /** Dependencies */
    import { Branch } from "@tripetto/vfsm/lib/branches/branch";
    import { IBranch } from "@tripetto/vfsm/lib/branches/interface";
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { StrictMatrix } from "@marblecore/matrix";
    export class Branches extends StrictMatrix<
        Branches,
        Branch,
        void,
        IBranch
    > {
        /**
         * Constructs a new branch dispatcher.
         * @param pState Reference to the parent state.
         */
        constructor(pState: State);
        /** Retrieves a reference to the parent state. */
        readonly state: State;
        /**
         * Creates a branch.
         * @param pBranch Reference to the branch ontology.
         * @param pAncestor Reference to the ancestor state.
         * @return Returns a reference to the branch instance.
         */
        create<T extends Branch>(pBranch: IBranch, pAncestor: State): T;
    }
}

declare module "@tripetto/vfsm/lib/branches/branch" {
    /** Dependencies */
    import { Branches } from "@tripetto/vfsm/lib/branches/branches";
    import { Callback } from "@marblecore/callback";
    import { Conditions } from "@tripetto/vfsm/lib/conditions/conditions";
    import { Context } from "@tripetto/vfsm/lib/instance/data/context";
    import { Factory } from "@tripetto/vfsm/lib/factory";
    import { IBranch } from "@tripetto/vfsm/lib/branches/interface";
    import { TBranchEvents } from "@tripetto/vfsm/lib/branches/events";
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { TBranchCulling } from "@tripetto/vfsm/lib/branches/culling";
    export abstract class Branch<T extends IBranch = IBranch> extends Factory<
        Branches,
        T,
        TBranchEvents
    > {
        /**
         * Constructs a new branch.
         * @param pBranches Reference to the branch dispatcher.
         * @param pBranch Reference to the branch data.
         */
        constructor(pBranches: Branches, pBranch: T);
        /** Retrieves a reference to the branches. */
        protected readonly branches: Branches;
        /** Retrieves an array with states. */
        readonly states: State[];
        /** Retrieves a reference to the parent state. */
        readonly state: State;
        /** Retrieves a reference to the condition dispatcher. */
        readonly conditions: Conditions;
        /** Retrieves the branch properties. */
        readonly props: Readonly<T>;
        /** Retrieves the culling mode for the branch. */
        readonly culling: TBranchCulling;
        /** Retrieves if this branch should spawn a new stack. */
        readonly spawn: boolean;
        /** Retrieves if this is the main branch. */
        readonly isMain: boolean;
        /**
         * Retrieves the entry state for the branch.
         * @param sMode Specifies the mode.
         * @return Returns the entry state.
         */
        entry(sMode: "live" | "preview"): State | undefined;
        /**
         * Invoked when the VFSM wants to verify the condition for the branch.
         * @param pContext Reference to the context.
         * @param pCallback Reference to the callback pointer. Pass this pointer as
         * return value to enable the asynchronous condition callback.
         * @return Returns the result of the condition or the callback reference if
         * the condition should be handled asynchronously.
         * @event
         */
        abstract onCondition(
            pContext: Context,
            pCallback: Callback<boolean>
        ): boolean | Callback<boolean>;
        /**
         * Invoked when the branch is ready for entering. The condition is approved.
         * @param pContext Reference to the context.
         * @event
         */
        onReady(pContext: Context): void;
        /**
         * Invoked when the branch is entered.
         * @param pContext Reference to the context.
         * @event
         */
        onEnter(pContext: Context): void;
        /**
         * Invoked when the branch is skipped due to a failing enter condition.
         * @param pContext Reference to the context.
         * @event
         */
        onSkip(pContext: Context): void;
        /**
         * Invoked when the branch is left.
         * @param pContext Reference to the context.
         * @event
         */
        onLeave(pContext: Context): void;
        /**
         * Invoked when the branch is canceled.
         * @param pContext Reference to the context.
         * @event
         */
        onCancel(pContext: Context): void;
    }
}

declare module "@tripetto/vfsm/lib/states/states" {
    /** Dependencies */
    import { Branch } from "@tripetto/vfsm/lib/branches/branch";
    import { IState } from "@tripetto/vfsm/lib/states/interface";
    import { Main } from "@tripetto/vfsm/lib/states/main";
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { StrictMatrix } from "@marblecore/matrix";
    import { TOntology } from "@tripetto/vfsm/lib/ontology";
    export class States extends StrictMatrix<States, State, void, IState> {
        /**
         * Constructs a new state dispatcher.
         * @param pOntology Reference to the ontology.
         */
        constructor(pOntology: TOntology);
        /** Retrieves a reference to the ontology. */
        readonly ontology: TOntology;
        /**
         * Creates a new main state.
         * @return Returns the main state.
         */
        main(): Main;
        /**
         * Creates a state.
         * @param pState Reference to the state data.
         * @param pBranch Reference to the parent branch.
         * @param pAncestor Reference to the ancestor state.
         * @param nNumerator Specifies the numerator of the state.
         * @return Returns a reference to the state.
         */
        create<T extends State>(
            pState: IState,
            pBranch: Branch,
            pAncestor: State,
            nNumerator: number
        ): T;
    }
}

declare module "@tripetto/vfsm/lib/states/state" {
    /** Dependencies */
    import { Await, Callback, ICallbackPromise } from "@marblecore/callback";
    import { Factory } from "@tripetto/vfsm/lib/factory";
    import { Branches } from "@tripetto/vfsm/lib/branches/branches";
    import { Transducers } from "@tripetto/vfsm/lib/transducers/transducers";
    import { States } from "@tripetto/vfsm/lib/states/states";
    import { Context } from "@tripetto/vfsm/lib/instance/data/context";
    import { TOntology } from "@tripetto/vfsm/lib/ontology";
    import { IState } from "@tripetto/vfsm/lib/states/interface";
    import { TStateEvents } from "@tripetto/vfsm/lib/states/events";
    export abstract class State<T extends IState = IState> extends Factory<
        States,
        T,
        TStateEvents
    > {
        /**
         * Contains the state numerator. This is a chronological number for each
         * state. The first state has a numerator value of `1` and the last state
         * equals the value of the `metrics.states` of a ontology instance. You can
         * use this property to calculate a simple progress indicator.
         */
        readonly numerator: number;
        /** Contains the await pointer for the state. */
        await: Await | undefined;
        /**
         * Constructs a new state.
         * @param pStates Reference to the state dispatcher.
         * @param pState Specifies the state data.
         */
        constructor(pStates: States, pState: T);
        /** Retrieves a reference to the states. */
        readonly states: States;
        /** Retrieves a reference to the ontology. */
        readonly ontology: TOntology;
        /** Retrieves a reference to the transducer dispatcher. */
        readonly transducers: Transducers;
        /** Retrieves a reference to the branch dispatcher. */
        readonly branches: Branches;
        /** Retrieves if the state is the main state. */
        readonly isMain: boolean;
        /** Retrieves the state properties. */
        readonly props: Readonly<T>;
        /** Retrieves if the state is evaluating. */
        readonly isEvaluating: boolean;
        /**
         * Evaluates the state.
         * @param pReference Reference to the callback or promise.
         * @param pContext Reference to the context.
         */
        evaluate<C>(
            pReference: Callback<C> | ICallbackPromise<C>,
            pContext: Context,
            fnAssert: (pPromise: ICallbackPromise<C>) => boolean
        ): void;
        /**
         * Invoked when the VFSM wants to verify the enter condition for the state.
         * This event is invoked just before the state is entered and it is also
         * invoked for the main state. This allows you to use a condition for the
         * main state which lets you control the main application branch. If you
         * override this function you have to make sure to return an appropriate
         * value for the main state.
         * @param pContext Reference to the context.
         * @param pCallback Reference to the callback pointer. Pass this pointer as
         * return value to enable the asynchronous condition callback.
         * @return Returns the result of the condition or the callback reference if
         * the condition should be handled asynchronously.
         * @event
         */
        abstract onEnterCondition(
            pContext: Context,
            pCallback: Callback<boolean>
        ): boolean | Callback<boolean>;
        /**
         * Invoked when state is entered.
         * @param pContext Reference to the context.
         * @param pAwait Reference to the await pointer. Pass this pointer as return
         * value to enable the asynchronous callback.
         * @return Returns the callback reference if the function should be handled
         * asynchronously.
         * @event
         */
        abstract onEnter(pContext: Context, pAwait: Await): void | Await;
        /**
         * Invoked when state is skipped due to a failing enter condition.
         * @param pContext Reference to the context.
         * @event
         */
        onSkip(pContext: Context): void;
        /**
         * Invoked when the VFSM wants to verify the leave condition for the state.
         * This event is invoked just before the state is left.
         * @param pContext Reference to the context.
         * @param pCallback Reference to the callback pointer. Pass this pointer as
         * return value to enable the asynchronous condition callback.
         * @return Returns the result of the condition or the callback reference if
         * the condition should be handled asynchronously.
         * @event
         */
        abstract onLeaveCondition(
            pContext: Context,
            pCallback: Callback<boolean>
        ): boolean | Callback<boolean>;
        /**
         * Invoked when the state is left.
         * @param pContext Reference to the context.
         * @return Returns the callback reference if the function should be handled
         * asynchronously.
         * @event
         */
        onLeave(pContext: Context): void;
        /**
         * Invoked when the state cannot be left due to a failing leave condition.
         * @param pContext Reference to the context.
         * @param pAwait Reference to the await pointer. Pass this pointer as return
         * value to enable the asynchronous callback.
         * @event
         */
        onLeaveError(pContext: Context, pAwait: Await): void | Await;
        /**
         * Invoked when the state is canceled.
         * @param pContext Reference to the context.
         * @event
         */
        onCancel(pContext: Context): void;
        /**
         * Invoked when the state needs to be validated.
         * @param pContext Reference to the context.
         * @param sCurrent Specifies the current validation state.
         * @param pCallback Reference to the callback pointer. Pass this pointer as
         * return value to enable the asynchronous validation callback.
         * @return Returns the result of the validation or the callback reference if
         * the validation should be handled asynchronously.
         * @event
         */
        abstract onValidate(
            pContext: Context,
            sCurrent: "unknown" | "fail" | "pass",
            pCallback: Callback<boolean>
        ): boolean | Callback<boolean>;
        /**
         * Invoked when the state is validated.
         * @param pContext Reference to the context.
         * @param sResult Specifies the validation state.
         * @param sType Specifies if the validation type.
         * @event
         */
        onValidated(
            pContext: Context,
            sResult: "fail" | "pass",
            sType: "initial" | "revalidate"
        ): void;
        /**
         * Invoked when the state starts or stops evaluating.
         * @param pContext Reference to the context.
         * @param bEvaluating Specifies if the state is evaluating.
         * @event
         */
        onEvaluate(pContext: Context, bEvaluating: boolean): void;
    }
}

declare module "@tripetto/vfsm/lib/transducers/transducers" {
    /** Dependencies */
    import { StrictMatrix } from "@marblecore/matrix";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { ITransducer } from "@tripetto/vfsm/lib/transducers/interface";
    export class Transducers extends StrictMatrix<
        Transducers,
        Transducer,
        void,
        ITransducer
    > {
        /**
         * Constructs a new transducer dispatcher.
         * @param pState Reference to the state.
         */
        constructor(pState: State);
        /** Retrieves a reference to the parent state. */
        readonly state: State;
        /**
         * Creates a transducer.
         * @param pTransducer Reference to the transducer ontology.
         * @return Returns a reference to the transducer instance.
         */
        create<T extends Transducer>(pTransducer: ITransducer): T;
    }
}

declare module "@tripetto/vfsm/lib/transducers/transducer" {
    /** Dependencies */
    import { Await, Callback } from "@marblecore/callback";
    import { Slots } from "@tripetto/slots";
    import { Factory } from "@tripetto/vfsm/lib/factory";
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { Transducers } from "@tripetto/vfsm/lib/transducers/transducers";
    import { Context } from "@tripetto/vfsm/lib/instance/data/context";
    import { ITransducer } from "@tripetto/vfsm/lib/transducers/interface";
    import { TTransducerEvents } from "@tripetto/vfsm/lib/transducers/events";
    export abstract class Transducer<
        T extends ITransducer = ITransducer
    > extends Factory<Transducers, T, TTransducerEvents> {
        /** Contains the await pointer for the transducer. */
        await: Await | undefined;
        /**
         * Constructs a new transducer.
         * @param pTransducers Reference to the transducer dispatcher.
         * @param pTransducer Reference to the transducer data.
         */
        constructor(pTransducers: Transducers, pTransducer: T);
        /** Retrieves a reference to the transducers. */
        protected readonly transducers: Transducers;
        /** Retrieves a reference to the parent state. */
        readonly state: State;
        /** Retrieves a reference to the slots. */
        readonly slots: Slots;
        /** Retrieves the transducer properties. */
        readonly props: Readonly<T>;
        /**
         * Seal or unseal the values.
         * @param pContext Reference to the context.
         * @param bSeal Specifies if the values should be sealed.
         */
        seal(pContext: Context, bSeal: boolean): void;
        /**
         * Retrieves if the seal should be verified.
         * @param pContext Reference to the context.
         * @return Returns `true` if the seal needs verification.
         */
        shouldVerifySeal(pContext: Context): boolean;
        /**
         * Invoked when the VFSM wants to verify the condition for the transducer.
         * This event is invoked just before the transducer is entered.
         * @param pContext Reference to the context.
         * @param pCallback Reference to the callback pointer. Pass this pointer as
         * return value to enable the asynchronous condition callback.
         * @return Returns the result of the condition or the callback reference if
         * the condition should be handled asynchronously.
         * @event
         */
        abstract onEnterCondition(
            pContext: Context,
            pCallback: Callback<boolean>
        ): boolean | Callback<boolean>;
        /**
         * Invoked when the transducer is entered.
         * @param pContext Reference to the context.
         * @param pAwait Reference to the await pointer. Pass this pointer as return
         * value to enable asynchronous validation.
         * @return Return the await pointer to implement asynchronous transducer validation.
         * @event
         */
        abstract onEnter(pContext: Context, pAwait: Await): void | Await;
        /**
         * Invoked when the transducer is skipped due to a failing enter condition.
         * @param pContext Reference to the context.
         * @event
         */
        onSkip(pContext: Context): void;
        /**
         * Invoked when the transducer is left.
         * @param pContext Reference to the context.
         * @event
         */
        onLeave(pContext: Context): void;
        /**
         * Invoked when the transducer is canceled.
         * @param pContext Reference to the context.
         * @event
         */
        onCancel(pContext: Context): void;
        /**
         * Invoked when the transducer is validated.
         * @param pContext Reference to the context.
         * @param sCurrent Specifies the current transducer validation state.
         * @param pCallback Reference to the callback pointer. Pass this pointer as
         * return value to enable the asynchronous validation callback.
         * @return Returns the result of the validation or the callback reference if
         * the validation should be handled asynchronously.
         * @event
         */
        abstract onValidate(
            pContext: Context,
            sCurrent: "unknown" | "fail" | "pass",
            pCallback: Callback<boolean>
        ): boolean | Callback<boolean>;
        /**
         * Invoked when the transducer is validated.
         * @param pContext Reference to the context.
         * @param sResult Specifies the validation result.
         * @param sType Specifies the validation type.
         * @event
         */
        onValidated(
            pContext: Context,
            sResult: "fail" | "pass",
            sType: "initial" | "revalidate"
        ): void;
    }
}

declare module "@tripetto/vfsm/lib/branches/interface" {
    /** Dependencies */
    import { IOntologyProperties } from "@tripetto/vfsm/lib/interface";
    import { TBranchCulling } from "@tripetto/vfsm/lib/branches/culling";
    export interface IBranch extends IOntologyProperties {
        /**
         * Specifies the jump reference for the branch. If this property is omitted
         * or an empty string, the branch successor will be determined
         * automatically. If this property is set to 'abort' the stack pointer will
         * be killed after the branch. If this property is set to the identifier
         * of another state, the stack pointer is redirected to that state.
         */
        readonly jump?: string | "abort";
        /** Specifies the culling mode for the branch. */
        readonly culling?: TBranchCulling;
        /** Specifies if this branch should spawn a new stack. */
        readonly spawn?: boolean;
    }
}

declare module "@tripetto/vfsm/lib/chain" {
    /** Dependencies */
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    import { Branch } from "@tripetto/vfsm/lib/branches/branch";
    import { Condition } from "@tripetto/vfsm/lib/conditions/condition";
    import { Slot } from "@tripetto/slots";
    export interface IChain {
        readonly states: {
            readonly [hash: string]: State | undefined;
        };
        readonly transducers: {
            readonly [hash: string]: Transducer | undefined;
        };
        readonly branches: {
            readonly [hash: string]: Branch | undefined;
        };
        readonly conditions: {
            readonly [hash: string]: Condition | undefined;
        };
        readonly slots: {
            readonly [hash: string]: Slot[] | Slot | undefined;
        };
    }
    export interface IChainLookup {
        /** Retrieves a state. */
        readonly state: (sHash: string) => State | undefined;
        /** Retrieves a transducer. */
        readonly transducer: (sHash: string) => Transducer | undefined;
        /** Retrieves a branch. */
        readonly branch: (sHash: string) => Branch | undefined;
        /** Retrieves a condition. */
        readonly condition: (sHash: string) => Condition | undefined;
        /** Retrieves a slot or slot group. */
        readonly slot: (sHash: string) => Slot | Slot[] | undefined;
    }
}

declare module "@tripetto/vfsm/lib/conditions/interface" {
    /** Dependencies */
    import { IOntologyProperties } from "@tripetto/vfsm/lib/interface";
    export interface ICondition extends IOntologyProperties {}
}

declare module "@tripetto/vfsm/lib/instance/data/interface" {
    /** Dependencies */
    import { TSerializeTypes } from "@marblecore/serializer";
    export interface IDataValue {
        /** Contains the value. */
        readonly value: TSerializeTypes;
        /** Contains the reference. */
        readonly reference: string | undefined;
        /** Contains the UTC set time of the value. */
        readonly time: number;
    }
    export interface IDataValues {
        readonly [context: string]: IDataValue | undefined;
    }
    export interface IDataImport {
        readonly [context: string]: Partial<IDataValue> | undefined;
    }
}

declare module "@tripetto/vfsm/lib/metrics" {
    export interface IMetrics {
        /** Number of states. */
        states: number;
        /** Number of transducers. */
        transducers: number;
        /** Number of branches. */
        branches: number;
        /** Number of conditions. */
        conditions: number;
        /** Number of slots. */
        slots: number;
    }
}

declare module "@tripetto/vfsm/lib/instance/data/immutable" {
    /** Dependencies */
    import { Slot } from "@tripetto/slots";
    import { THookContext } from "@marblecore/hookup";
    import { TSerializeTypes } from "@marblecore/serializer";
    import { Value } from "@tripetto/vfsm/lib/instance/data/value";
    export class ImmutableValue<
        T extends TSerializeTypes = TSerializeTypes,
        S extends Slot<T> = Slot<T>
    > {
        /**
         * Creates an immutable value instance.
         * @param pValue Reference to the mutable value instance.
         */
        constructor(pValue: Value<T, S>);
        /** Retrieves a unique key for the value within its context (the key is prefixed with an underscore). */
        readonly key: string;
        /** Retrieves the context for the value. */
        readonly context: "*" | string;
        /** Retrieves a reference to the slot. */
        readonly slot: S;
        /** Retrieves the value. */
        readonly value: T;
        /**
         * Retrieves the value as a string (this will always return a string with a
         * value regardless of a pristine value was set).
         */
        readonly string: string;
        /** Retrieves the optional value reference. */
        readonly reference: string | undefined;
        /** Retrieves the value set timestamp. */
        readonly time: number | undefined;
        /** Retrieves if there is a value. */
        readonly hasValue: boolean;
        /** Retrieves if the value is modified. */
        readonly isModified: boolean;
        /** Retrieves if the value is sealed. */
        readonly isSealed: boolean;
        /** Retrieves if the value is locked. */
        readonly isLocked: boolean;
        /**
         * Subscribe to value changes using the given context.
         * @param fnChange Specifies the event that is invoked when the value is changed.
         * @param pContext Specifies the context.
         * @return Returns a reference to the instance.
         */
        subscribe(
            fnChange: (pValue: ImmutableValue<T, S>) => void,
            pContext?: THookContext
        ): this;
        /**
         * Unsubscribe the given context from the value.
         * @param pContext Specifies the context.
         * @return Returns a reference to the instance.
         */
        unsubscribe(pContext?: THookContext): this;
    }
}

declare module "@tripetto/vfsm/lib/instance/moments/moment" {
    /** Dependencies */
    import { Context } from "@tripetto/vfsm/lib/instance/data/context";
    import { IMomentSnapshot } from "@tripetto/vfsm/lib/instance/moments/snapshot";
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    export interface IMoment {
        /** Index of the moment */
        readonly index: number;
        /** Snapshot of the moment. */
        readonly snapshot: IMomentSnapshot;
        /** Reference to the state. */
        readonly state: State;
        /** Reference to the validated transducers. */
        readonly transducers: Transducer[];
        /** Context for the data. */
        readonly context: Context;
        /** Specifies the tense of the moment. */
        readonly tense: "past" | "present" | "future";
        /** Restores the moment. */
        readonly restore: () => void;
    }
    export interface IMoments {
        readonly [id: number]: IMoment | undefined;
    }
}

declare module "@tripetto/vfsm/lib/instance/instance" {
    import { Data } from "@tripetto/vfsm/lib/instance/data/data";
    import { IChainLookup } from "@tripetto/vfsm/lib/chain";
    import { TInstanceEvents } from "@tripetto/vfsm/lib/instance/events";
    import { Condition } from "@tripetto/vfsm/lib/conditions/condition";
    import { IInstanceOrchestrator } from "@tripetto/vfsm/lib/instance/orchestrator";
    import { IInstanceSnapshot } from "@tripetto/vfsm/lib/instance/snapshot";
    import { Item } from "@marblecore/matrix";
    import { IMoment } from "@tripetto/vfsm/lib/instance/moments/moment";
    import { Slot } from "@tripetto/slots";
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    import { TThreading } from "@tripetto/vfsm/lib/threading";
    export class Instance extends Item<TInstanceEvents> {
        /**
         * Constructs the instance.
         * @param pOrchestrator Specifies the instance orchestrator.
         * @param sThreading Specifies the threading model (`single` or `multi`).
         * @param pBootstrap Specifies the bootstrap state or snapshot.
         */
        constructor(
            pOrchestrator: IInstanceOrchestrator,
            sThreading: TThreading,
            pBootstrap?: State | IInstanceSnapshot
        );
        /** Retrieves the instance identifier. */
        readonly id: string;
        /** Retrieves the unique ontology fingerprint. */
        readonly fingerprint: string;
        /** Retrieves if the instance allows concurrency. */
        readonly allowConcurrency: boolean;
        /** Retrieves if the preview mode is enabled. */
        readonly isPreview: boolean;
        /** Retrieves if the instance is running. */
        readonly isRunning: boolean;
        /** Retrieves if the instance is stopped (interrupted). */
        readonly isStopped: boolean;
        /** Retrieves if the instance is paused. */
        readonly isPaused: boolean;
        /** Retrieves if the instance has finished. */
        readonly isFinished: boolean;
        /** Retrieves if the instance is at the start. */
        readonly isAtStart: boolean;
        /** Retrieves if the instance is at the head. */
        readonly isAtHead: boolean;
        /** Retrieves if the instance is at the finish. */
        readonly isAtFinish: boolean;
        /**
         * Retrieves if the instance is finishable using the `finish()` method.
         * Only instances that have collected sealed values in one or more slots are finishable.
         */
        readonly isFinishable: boolean;
        /** Retrieves the chain lookup object. */
        readonly lookup: IChainLookup;
        /** Retrieves the data. */
        readonly data: Data;
        /** Retrieves the exportable slots. */
        readonly slots: Slot[];
        /** Retrieves a history of all moments. */
        readonly history: IMoment[];
        /** Retrieves the checksum. */
        readonly checksum: string;
        /** Retrieves the present moment. */
        readonly presentMoment: IMoment | undefined;
        /** Retrieves the instance snapshot. */
        readonly snapshot: IInstanceSnapshot;
        /** Retrieves instance error information. */
        readonly errors: number;
        /**
         * Starts the instance.
         * @return Returns `true` if the instance is started.
         */
        start(): boolean;
        /** Restarts the instance. */
        restart(): void;
        /**
         * Stops the instance.
         * @return Returns `true` if the instance was running and is stopped.
         */
        stop(): boolean;
        /**
         * Pauses the instance.
         * @return Returns the instance snapshot necessary to resume the instance.
         */
        pause(): IInstanceSnapshot;
        /**
         * Step forward in time.
         * @return Returns `true` if the step succeeded.
         */
        stepForward(): boolean;
        /**
         * Step backward in time.
         * @return Returns `true` if the step succeeded.
         */
        stepBackward(): boolean;
        /**
         * Step to the start.
         * @return Returns `true` if the step succeeded.
         */
        stepToStart(): boolean;
        /**
         * Step to the head of the history.
         * @return Returns `true` if the step succeeded.
         */
        stepToHead(): boolean;
        /**
         * Finish the instance when it is at the end and there are slots collected.
         * @return Returns `true` if the instance was finished.
         */
        finish(): boolean;
        /**
         * Iterates through all possible transducers and their context.
         * @param fnTransducer Specifies a function that is invoked for each transducer.
         */
        transducers<T extends Transducer, C extends Condition>(
            fnTransducer: (pTransducer: T, pContext: C[]) => void
        ): void;
    }
}

declare module "@tripetto/vfsm/lib/interface" {
    /** Dependencies */
    import { ISerialize } from "@marblecore/serializer";
    import { IBranch } from "@tripetto/vfsm/lib/branches/interface";
    export interface IOntologyProperties extends ISerialize {
        /** Contains the property identifier. */
        readonly id: string;
    }
    export interface IOntology extends IBranch {}
}

declare module "@tripetto/vfsm/lib/instance/snapshot" {
    /** Dependencies */
    import { IDataSnapshot } from "@tripetto/vfsm/lib/instance/data/snapshot";
    import { IMomentSnapshot } from "@tripetto/vfsm/lib/instance/moments/snapshot";
    export interface IInstanceSnapshot {
        /** Contains the ontology hash. */
        readonly a: string;
        /** Contains the moment snapshots. */
        readonly b: IMomentSnapshot[];
        /** Contains the data snapshots. */
        readonly c?: IDataSnapshot[];
        /** Contains the index of the active snapshot. */
        readonly d?: number;
        /** Contains if the instance allows concurrency. */
        readonly e?: true;
    }
}

declare module "@tripetto/vfsm/lib/snapshot" {
    /** Dependencies */
    import { IInstanceSnapshot } from "@tripetto/vfsm/lib/instance/snapshot";
    export interface ISnapshot {
        /** Contains the instance data. */
        readonly a: IInstanceSnapshot[];
    }
}

declare module "@tripetto/vfsm/lib/states/interface" {
    /** Dependencies */
    import { IOntologyProperties } from "@tripetto/vfsm/lib/interface";
    export interface IState extends IOntologyProperties {}
}

declare module "@tripetto/vfsm/lib/transducers/interface" {
    /** Dependencies */
    import { ISlot } from "@tripetto/slots";
    import { IOntologyProperties } from "@tripetto/vfsm/lib/interface";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    import { Condition } from "@tripetto/vfsm/lib/conditions/condition";
    export interface ITransducer extends IOntologyProperties {
        /** Contains the slots for the transducer. */
        readonly slots?: ISlot[];
    }
    export interface ITransducers {
        /** Conditions for the transducers. */
        conditions: Condition[];
        /** Transducers. */
        transducers: (ITransducers | Transducer)[];
    }
}

declare module "@tripetto/vfsm/lib/instance/data/variable" {
    /** Dependencies */
    import { Slot } from "@tripetto/slots";
    import { THookContext } from "@marblecore/hookup";
    import { TSerializeTypes } from "@marblecore/serializer";
    export interface IVariable {
        /** Contains the value as a string. */
        readonly string: string;
        /** Reference to the slot. */
        readonly slot?: Slot;
        /** Contains the value. */
        readonly value?: TSerializeTypes;
        /** Subscribe to variable changes. */
        readonly subscribe: (
            fnChange: (pVariable: IVariable) => void,
            pContext?: THookContext
        ) => void;
        /** Unsubscribe a context from variable changes. */
        readonly unsubscribe: (pContext: THookContext) => void;
    }
}

declare module "@tripetto/vfsm/lib/instance/data/context" {
    /** Dependencies */
    import { Slot } from "@tripetto/slots";
    import { Condition } from "@tripetto/vfsm/lib/conditions/condition";
    import { IChainLookup } from "@tripetto/vfsm/lib/chain";
    import { ImmutableValue } from "@tripetto/vfsm/lib/instance/data/immutable";
    import { Instance } from "@tripetto/vfsm/lib/instance/instance";
    import { IVariable } from "@tripetto/vfsm/lib/instance/data/variable";
    import { Stack } from "@tripetto/vfsm/lib/instance/stack/stack";
    import { TSerializeTypes } from "@marblecore/serializer";
    import { Value } from "@tripetto/vfsm/lib/instance/data/value";
    export interface IContext {
        /** Reference to the condition. */
        readonly condition: Condition;
        /** Specifies the context. */
        readonly context: string;
    }
    export class Context {
        /**
         * Creates a new context.
         * @param pStack Reference to the stack.
         * @param pChain Reference to the chain.
         */
        constructor(pStack: Stack, pChain: IChainLookup);
        /** Retrieves the instance of the context. */
        readonly instance: Instance;
        /** Retrieves the context key. */
        readonly key: string;
        /**
         * Retrieves if this context is the mutable context of the supplied stack.
         * @param pStack Reference to the stack.
         * @return Returns `true` if the context is of the supplied stack.
         */
        isMutableContextOf(pStack: Stack): boolean;
        /**
         * Retrieves the mutable value for the supplied slot instance or identifier.
         * @param pSlotOrId Reference to the slot or slot identifier.
         * @return Returns the value or `undefined` if no value was found.
         */
        mutableValueFor<T extends TSerializeTypes>(
            pSlotOrId: Slot | string
        ): Value<T> | undefined;
        /**
         * Retrieves the immutable value for the supplied slot instance or identifier.
         * @param pSlotOrId Reference to the slot or slot identifier.
         * @return Returns the value or `undefined` if no value was found.
         */
        immutableValueFor<T extends TSerializeTypes>(
            pSlotOrId: Slot | string
        ): ImmutableValue<T> | undefined;
        /**
         * Retrieves the variable for the supplied slot instance or (group) identifier.
         * @param pSlotOrId Reference to the slot instance or (group) identifier.
         * @param fnGroupId Specifies the function used to retrieve the group identifier.
         * @return Returns the variable or `undefined` if the variable was not found,
         * not valid or not set.
         */
        variableFor(
            pSlotOrId: Slot | string,
            fnGroupId?: (pCondition: Condition) => string | undefined
        ): IVariable | undefined;
    }
}

declare module "@tripetto/vfsm/lib/instance/data/data" {
    import { Matrix } from "@marblecore/matrix";
    import { Slot } from "@tripetto/slots";
    import { States } from "@tripetto/vfsm/lib/states/states";
    import { TSerializeTypes } from "@marblecore/serializer";
    import { TValuesEvents } from "@tripetto/vfsm/lib/instance/data/events";
    import { Values } from "@tripetto/vfsm/lib/instance/data/values";
    export class Data extends Matrix<Values, TValuesEvents> {
        /**
         * Creates a new data instance.
         * @param pStates Reference to the states.
         */
        constructor(pStates: States);
        /** Retrieves all exportable slots. */
        readonly slots: Slot[];
        /** Retrieves all exportable values. */
        readonly all: Values[];
        /** Retrieves if there are sealed values collected. */
        readonly hasSealedValues: boolean;
        /**
         * Retrieve the values for the supplied slot.
         * @param pSlotOrId Reference to the slot or slot identifier.
         * @return Returns the values or `undefined` if no values are found.
         */
        forSlot<T extends TSerializeTypes, S extends Slot<T>>(
            pSlotOrId: Slot | string
        ): Values<T, S> | undefined;
    }
}

declare module "@tripetto/vfsm/lib/ontology" {
    /** Dependencies */
    import { Branch } from "@tripetto/vfsm/lib/branches/branch";
    import { ISnapshot } from "@tripetto/vfsm/lib/snapshot";
    import { Condition } from "@tripetto/vfsm/lib/conditions/condition";
    import { Errors } from "@tripetto/vfsm/enum/errors";
    import { IConstructors } from "@tripetto/vfsm/lib/constructors";
    import { IInstanceSnapshot } from "@tripetto/vfsm/lib/instance/snapshot";
    import { Matrix } from "@marblecore/matrix";
    import { IMetrics } from "@tripetto/vfsm/lib/metrics";
    import { Instance } from "@tripetto/vfsm/lib/instance/instance";
    import { TOntologyEvents } from "@tripetto/vfsm/lib/events";
    import { Main } from "@tripetto/vfsm/lib/states/main";
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { States } from "@tripetto/vfsm/lib/states/states";
    import { TBranch } from "@tripetto/vfsm/lib/branches/type";
    import { TCondition } from "@tripetto/vfsm/lib/conditions/type";
    import { THooks } from "@marblecore/hookup";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    import { TState } from "@tripetto/vfsm/lib/states/type";
    import { TThreading } from "@tripetto/vfsm/lib/threading";
    import { TTransducer } from "@tripetto/vfsm/lib/transducers/type";
    export type TOntology = Ontology<State, Transducer, Branch, Condition>;
    export abstract class Ontology<
        S extends State,
        T extends Transducer,
        B extends Branch,
        C extends Condition,
        O extends {} = {},
        H extends THooks = void
    > extends Matrix<Instance, TOntologyEvents | H> {
        /**
         * Constructs a new ontology.
         * @param pState Reference to the state constructor.
         * @param pTransducer Reference to the transducer constructor.
         * @param pBranch Reference to the branch constructor.
         * @param pCondition Reference to the condition constructor.
         * @param pOntology Reference to the ontology data.
         * @param bAutoPushForward Specifies if auto push-forward should be used. when
         * enabled, steps forward are made automatically as long as there are no
         * validation obstructions.
         * @param bPreview Specifies if the preview mode is enabled. In this mode all
         * conditions and validations will pass without evaluating them. This allows
         * an application to iterate through all possible states and transducers.
         * @return Returns a reference to the instance.
         * @param bVerbose Enables or disables verbose logging.
         */
        constructor(
            pState: TState<S>,
            pTransducer: TTransducer<T>,
            pBranch: TBranch<B>,
            pCondition: TCondition<C>,
            pOntology?: string | O,
            bAutoPushForward?: boolean,
            bPreview?: boolean,
            bVerbose?: boolean
        );
        /** Retrieves the constructors. */
        readonly constructors: IConstructors<S, T, B, C>;
        /** Retrieves the ontology data. */
        readonly ontology: Readonly<O> | undefined;
        /** Retrieves the unique ontology fingerprint. */
        readonly fingerprint: string;
        /** Retrieves the ontology snapshot. */
        readonly snapshot: ISnapshot;
        /** Retrieves a reference to the state dispatcher. */
        readonly states: States;
        /** Retrieves a reference to the main state. */
        readonly main: Main | undefined;
        /** Retrieves if the ontology is loaded successfully. */
        readonly isLoaded: boolean;
        /** Retrieves if the VFSM is running. */
        readonly isRunning: boolean;
        /** Retrieves if the VFSM is stopped. */
        readonly isStopped: boolean;
        /** Retrieves if the VFSM has finished (without pausing or stopping it). */
        readonly isFinished: boolean;
        /** Retrieves if the VFSM is paused. */
        readonly isPaused: boolean;
        /** Retrieves if the preview mode is enabled. */
        readonly isPreview: boolean;
        /** Enables or disables verbose logging. */
        isVerbose: boolean;
        /** Enables or disables automatic push-forward. */
        autoPushForward: boolean;
        /** Retrieves the ontology metrics. */
        readonly metrics: Readonly<IMetrics>;
        /** Retrieves the log message list. */
        readonly log: string[];
        /** Sets the log function. */
        logger: (sMessage: string) => void;
        /**
         * Invoked when an instance is created.
         * @param pInstance Reference to the instance.
         * @event
         */
        protected onInstanceCreate(pInstance: Instance): void;
        /**
         * Invoked when an instance is started.
         * @param pInstance Reference to the instance.
         * @event
         */
        protected onInstanceStart(pInstance: Instance): void;
        /**
         * Invoked when an instance ends. This event is always invoked
         * when an instance is finished, stopped or paused.
         * @param pInstance Reference to the instance.
         * @param sType Specifies why the instance has ended. Can be one of the
         * following values:
         * - `finished`: The instance was finished;
         * - `stopped`: The instance was forced to stop;
         * - `paused`: The instance was paused.
         * @event
         */
        protected onInstanceEnd(
            pInstance: Instance,
            sType: "finished" | "stopped" | "paused"
        ): void;
        /**
         * Invoked when an instance has finished.
         * @param pInstance Reference to the instance.
         * @event
         */
        protected onInstanceFinish(pInstance: Instance): void;
        /**
         * Invoked when an instance is paused.
         * @param pInstance Reference to the instance.
         * @event
         */
        protected onInstancePause(pInstance: Instance): void;
        /**
         * Invoked when an instance is stopped.
         * @param pInstance Reference to the instance.
         * @event
         */
        protected onInstanceStop(pInstance: Instance): void;
        /**
         * Spawns a new instance. Throws an error if no ontology is loaded.
         * @param sThreading Specifies if a single or multi threaded instance should be spawned.
         * @param pSnapshot Specifies the optional instance snapshot.
         * @return Returns a reference to the instance.
         */
        protected spawn(
            sThreading?: TThreading,
            pSnapshot?: IInstanceSnapshot
        ): Instance;
        /**
         * Loads an ontology.
         * @param pOntology Specifies the ontology data.
         * @return Returns `Errors.None` if the ontology is loaded.
         */
        load(pOntology: O | string): Errors;
        /** Unloads the ontology. */
        unload(): void;
        /**
         * Starts a new instance. Throws an error if no ontology is loaded.
         * @param pSnapshot Specifies the optional instance snapshot.
         * @return Returns a reference to the instance.
         */
        start(pSnapshot?: IInstanceSnapshot): Instance;
        /**
         * Stops all running instances.
         * @return Returns `true` if there was an actual instance running.
         */
        stop(): boolean;
        /**
         * Pauses the VFSM ontology.
         * @return Returns the snapshot which is necessary to resume the VFSM
         * or `undefined` if no instance was running.
         */
        pause(): ISnapshot | undefined;
        /**
         * Resumes the VFSM ontology. This will resume all instances that are
         * present in the snapshot data.
         * @param pSnapshot Specifies the snapshot data.
         * @return Returns `true` if the resume is successful and the VFSM is
         * running.
         */
        resume(pSnapshot: ISnapshot): boolean;
        /**
         * Enables or disables the preview mode. Any running instances are updated.
         * @param bPreview Specifies if the preview is enabled or not.
         */
        previewMode(bEnabled: boolean): void;
        /**
         * Iterates through each item in the ontology.
         * @param fnItem Callback which is invoked for each item.
         * @param sType Specifies the type of items.
         * @return Returns `true` if the iteration was interrupted.
         */
        forEach<Types extends State | Transducer | Branch | Condition>(
            fnItem: (pItem: Types) => boolean | void,
            sType?:
                | "states"
                | "transducers"
                | "branches"
                | "conditions"
                | "all",
            pExclude?: State | Transducer | Branch | Condition
        ): boolean;
    }
}

declare module "@tripetto/vfsm/lib/instance/data/value" {
    import { THookContext } from "@marblecore/hookup";
    import { ImmutableValue } from "@tripetto/vfsm/lib/instance/data/immutable";
    import { Slot } from "@tripetto/slots";
    import { TSerializeTypes } from "@marblecore/serializer";
    export class Value<
        T extends TSerializeTypes = TSerializeTypes,
        S extends Slot<T> = Slot<T>
    > {
        /** Specifies if the value was locked (readonly). */
        isLocked: boolean;
        /**
         * Creates a new value store.
         * @param pSlot Reference to the slot.
         * @param sContext Contains the context of the value.
         * @param fnChange Contains the change callback.
         * @param pValue Contains the hot value.
         * @param sReference Contains the value reference.
         * @param nTime Contains the UTC set time of the value.
         * @param bSealed Specifies if the value was sealed.
         * @param bLocked Specifies if the value was locked.
         */
        constructor(
            pSlot: S,
            sContext: string,
            fnChange: (pValue: Value<T, S>) => void,
            pValue?: TSerializeTypes,
            sReference?: string,
            nTime?: number,
            bSealed?: boolean,
            bLocked?: boolean
        );
        /** Retrieves the value instance identifier. */
        readonly id: number;
        /** Retrieves a unique key for the value within its context (the key is prefixed with an underscore). */
        readonly key: string;
        /** Retrieves the context for the value. */
        readonly context: "*" | string;
        /** Retrieves a reference to the slot. */
        readonly slot: S;
        /** Sets the pristine value. */
        pristine: TSerializeTypes;
        /** Sets the value. */
        value: T;
        /**
         * Retrieves the value as a string (this will always return a string with a
         * value regardless of a pristine value was set).
         */
        readonly string: string;
        /** Retrieves the optional value reference. */
        readonly reference: string | undefined;
        /** Retrieves if there is a value. */
        readonly hasValue: boolean;
        /** Retrieves if the value is modified. */
        readonly isModified: boolean;
        /** Specifies if the value is sealed. */
        isSealed: boolean;
        /** Specifies if the value is dirty. */
        isDirty: boolean;
        /** Retrieves if the seal should to be verified. */
        readonly shouldVerifySeal: boolean;
        /** Retrieves the value set timestamp. */
        readonly time: number | undefined;
        /** Retrieves a immutable representation of the value. */
        readonly immutable: ImmutableValue<T, S>;
        /**
         * Sets the pristine value and optional reference.
         * @param pValue Specifies the pristine value.
         * @param sReference Specifies an optional value reference.
         * @return Retrieves the value.
         */
        set(pValue: TSerializeTypes, sReference?: string): T;
        /**
         * Imports a value.
         * @param pValue Specifies the pristine value.
         * @param sReference Specifies an optional value reference.
         * @param nTime Specifies the set timestamp.
         * @return Retrieves the value.
         */
        import(pValue: TSerializeTypes, sReference?: string, nTime?: number): T;
        /**
         * Marks the value as confirmed. In this case it is not necessary to
         * explicit set a value to make the value valid.
         * @return Returns a reference to the instance.
         */
        confirm(): this;
        /**
         * Marks the value as dirty. An explicit value should be set to mark
         * the value as valid or `confirm()` should be invoked.
         * @return Returns a reference to the instance.
         */
        dirty(): this;
        /**
         * Subscribe to value changes using the given context.
         * @param fnChange Specifies the event that is invoked when the value is changed.
         * @param pContext Specifies the context.
         * @return Returns a reference to the instance.
         */
        subscribe(
            fnChange: (pValue: Value<T, S>) => void,
            pContext?: THookContext
        ): this;
        /**
         * Unsubscribe the given context from the value.
         * @param pContext Specifies the context.
         * @return Returns a reference to the instance.
         */
        unsubscribe(pContext?: THookContext): this;
    }
}

declare module "@tripetto/vfsm/lib/instance/data/values" {
    import { Data } from "@tripetto/vfsm/lib/instance/data/data";
    import {
        IDataImport,
        IDataValue,
        IDataValues
    } from "@tripetto/vfsm/lib/instance/data/interface";
    import { IDataSnapshot } from "@tripetto/vfsm/lib/instance/data/snapshot";
    import { ImmutableValue } from "@tripetto/vfsm/lib/instance/data/immutable";
    import { Item } from "@marblecore/matrix";
    import { TValuesEvents } from "@tripetto/vfsm/lib/instance/data/events";
    import { Slot } from "@tripetto/slots";
    import { Stack } from "@tripetto/vfsm/lib/instance/stack/stack";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    import { TSerializeTypes } from "@marblecore/serializer";
    import { Value } from "@tripetto/vfsm/lib/instance/data/value";
    export class Values<
        T extends TSerializeTypes = TSerializeTypes,
        S extends Slot<T> = Slot<T>
    > extends Item<TValuesEvents> {
        /**
         * Constructs a new values store.
         * @param pTransducer Reference to the transducer.
         * @param pSlot Reference to the slot.
         */
        constructor(pTransducer: Transducer, pSlot: S);
        /** Retrieves a reference to the data instance. */
        readonly data: Data;
        /** Retrieves a reference to the transducer. */
        readonly transducer: Transducer;
        /** Retrieves a reference to the slot. */
        readonly slot: S;
        /** Retrieves the values. */
        readonly values: IDataValues;
        /** Retrieves if there are sealed values collected. */
        readonly hasSealedValues: boolean;
        /**
         * Retrieves the most relevant value for the supplied stack.
         * @param pStack Reference to the stack to use as a context.
         * @return Returns a mutable value.
         */
        mutableValue(pStack: Stack | undefined): Value<T, S>;
        /**
         * Retrieves the most relevant value for the supplied stack.
         * @param pStack Reference to the stack to use as a context.
         * @return Returns an immutable value.
         */
        immutableValue(
            pStack: Stack | undefined
        ): ImmutableValue<T, S> | undefined;
        /**
         * Retrieves if the value of the specified context was modified.
         * @param sContext Specifies the context.
         * @return Returns `true` if the value was modified.
         */
        isModified(sContext: string): boolean;
        /**
         * Snapshots the store.
         * @return Returns the store snapshot.
         */
        snapshot(): IDataSnapshot | undefined;
        /**
         * Restores the values.
         * @param pSnapshot Specifies the snapshot to restore.
         */
        restore(pSnapshot: IDataSnapshot): void;
        /**
         * Imports a single value.
         * @param sContext Specifies the context.
         * @param pData Specifies the data to import.
         */
        importValue(sContext: string, pData: Partial<IDataValue>): void;
        /**
         * Import a values object.
         * @param pValues Specifies the values to import.
         */
        import(pValues: IDataImport): void;
    }
}

declare module "@tripetto/vfsm/enum/errors" {
    export enum Errors {
        /** No errors were detected. */
        None = 0,
        /** The state data in the ontology is invalid. */
        OntologyInvalidState = 1,
        /** The transducer data in the ontology is invalid. */
        OntologyInvalidTransducer = 2,
        /** The branch data in the ontology is invalid. */
        OntologyInvalidBranch = 4,
        /** The condition data in the ontology is invalid. */
        OntologyInvalidCondition = 8,
        /** The jump reference in the ontology is invalid. */
        OntologyInvalidJump = 16,
        /** The resume data is invalid. */
        ResumeInvalidHash = 32,
        /** One of more resume data slots are invalid. */
        ResumeInvalidData = 64,
        /** The resume stack is invalid. */
        ResumeInvalidStack = 128
    }
}

declare module "@tripetto/vfsm/lib/decorator" {
    import { TBranch } from "@tripetto/vfsm/lib/branches/type";
    import { TCondition } from "@tripetto/vfsm/lib/conditions/type";
    import { TState } from "@tripetto/vfsm/lib/states/type";
    import { TTransducer } from "@tripetto/vfsm/lib/transducers/type";
    export type TTypes = TState | TTransducer | TBranch | TCondition;
    /**
     * Sets the property name for an ontology.
     * @param sProperty Specifies the property name.
     */
    export function ontology<T extends TTypes>(
        sProperty: string
    ): (pType: T) => void;
}

declare module "@tripetto/vfsm/lib/events" {
    /** Dependencies */
    import { IHookPayload, THooks } from "@marblecore/hookup";
    import { TOntology } from "@tripetto/vfsm/lib/ontology";
    import { Instance } from "@tripetto/vfsm/lib/instance/instance";
    export type TOntologyEvents =
        | "OnLoad"
        | "OnUnload"
        | "OnStart"
        | "OnEnd"
        | "OnPause"
        | "OnStop"
        | "OnFinish"
        | "OnInstanceCreate"
        | "OnInstanceStart"
        | "OnInstanceEnd"
        | "OnInstancePause"
        | "OnInstanceStop"
        | "OnInstanceFinish";
    export interface IOntologyEvent<T extends THooks> extends IHookPayload<T> {
        readonly ontology: TOntology;
    }
    export interface IOntologyLoadEvent extends IOntologyEvent<"OnLoad"> {}
    export interface IOntologyUnloadEvent extends IOntologyEvent<"OnUnload"> {}
    export interface IOntologyStartEvent extends IOntologyEvent<"OnStart"> {}
    export interface IOntologyPauseEvent extends IOntologyEvent<"OnPause"> {}
    export interface IOntologyStopEvent extends IOntologyEvent<"OnStop"> {}
    export interface IOntologyFinishEvent extends IOntologyEvent<"OnFinish"> {}
    export interface IOntologyEndEvent extends IOntologyEvent<"OnEnd"> {
        readonly type: "finished" | "stopped" | "paused";
    }
    export interface IOntologyInstanceEvent<T extends THooks>
        extends IOntologyEvent<T> {
        readonly instance: Instance;
    }
    export interface IOntologyInstanceCreateEvent
        extends IOntologyInstanceEvent<"OnInstanceCreate"> {}
    export interface IOntologyInstanceStartEvent
        extends IOntologyInstanceEvent<"OnInstanceStart"> {}
    export interface IOntologyInstancePauseEvent
        extends IOntologyInstanceEvent<"OnInstancePause"> {}
    export interface IOntologyInstanceStopEvent
        extends IOntologyInstanceEvent<"OnInstanceStop"> {}
    export interface IOntologyInstanceFinishEvent
        extends IOntologyInstanceEvent<"OnInstanceFinish"> {}
    export interface IOntologyInstanceEndEvent
        extends IOntologyInstanceEvent<"OnInstanceEnd"> {
        readonly type: "finished" | "stopped" | "paused";
    }
}

declare module "@tripetto/vfsm/lib/instance/events" {
    /** Dependencies */
    import { IHookPayload, THooks } from "@marblecore/hookup";
    import { Instance } from "@tripetto/vfsm/lib/instance/instance";
    export type TInstanceEvents =
        | "OnCreate"
        | "OnStart"
        | "OnEnd"
        | "OnPause"
        | "OnStop"
        | "OnFinish";
    export interface IInstanceEvent<T extends THooks> extends IHookPayload<T> {
        readonly instance: Instance;
    }
    export interface IInstanceCreateEvent extends IInstanceEvent<"OnCreate"> {}
    export interface IInstanceStartEvent extends IInstanceEvent<"OnStart"> {}
    export interface IInstancePauseEvent extends IInstanceEvent<"OnPause"> {}
    export interface IInstanceStopEvent extends IInstanceEvent<"OnStop"> {}
    export interface IInstanceFinishEvent extends IInstanceEvent<"OnFinish"> {}
    export interface IInstanceEndEvent extends IInstanceEvent<"OnEnd"> {
        readonly type: "finished" | "stopped" | "paused";
    }
}

declare module "@tripetto/vfsm/lib/branches/events" {
    /** Dependencies */
    import { IHookPayload, THooks } from "@marblecore/hookup";
    import { Context } from "@tripetto/vfsm/lib/instance/data/context";
    import { Branch } from "@tripetto/vfsm/lib/branches/branch";
    export type TBranchEvents =
        | "OnReady"
        | "OnEnter"
        | "OnSkip"
        | "OnLeave"
        | "OnCancel";
    export interface IBranchEvent<T extends THooks> extends IHookPayload<T> {
        readonly branch: Branch;
        readonly context: Context;
    }
    export interface IBranchReadyEvent extends IBranchEvent<"OnReady"> {}
    export interface IBranchEnterEvent extends IBranchEvent<"OnEnter"> {}
    export interface IBranchSkipEvent extends IBranchEvent<"OnSkip"> {}
    export interface IBranchLeaveEvent extends IBranchEvent<"OnLeave"> {}
    export interface IBranchCancelEvent extends IBranchEvent<"OnCancel"> {}
}

declare module "@tripetto/vfsm/lib/states/events" {
    /** Dependencies */
    import { IHookPayload, THooks } from "@marblecore/hookup";
    import { Context } from "@tripetto/vfsm/lib/instance/data/context";
    import { State } from "@tripetto/vfsm/lib/states/state";
    export type TStateEvents =
        | "OnEnter"
        | "OnSkip"
        | "OnLeave"
        | "OnCancel"
        | "OnValidated"
        | "OnEvaluate";
    export interface IStateEvent<T extends THooks> extends IHookPayload<T> {
        readonly state: State;
        readonly context: Context;
    }
    export interface IStateEnterEvent extends IStateEvent<"OnEnter"> {}
    export interface IStateSkipEvent extends IStateEvent<"OnSkip"> {}
    export interface IStateLeaveEvent extends IStateEvent<"OnLeave"> {}
    export interface IStateCancelEvent extends IStateEvent<"OnCancel"> {}
    export interface IStateEvaluateEvent extends IStateEvent<"OnEvaluate"> {
        readonly evaluating: boolean;
    }
    export interface IStateValidatedEvent extends IStateEvent<"OnValidated"> {
        readonly result: "fail" | "pass";
        readonly type: "initial" | "revalidate";
    }
}

declare module "@tripetto/vfsm/lib/transducers/events" {
    /** Dependencies */
    import { IHookPayload, THooks } from "@marblecore/hookup";
    import { Context } from "@tripetto/vfsm/lib/instance/data/context";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    export type TTransducerEvents =
        | "OnEnter"
        | "OnSkip"
        | "OnLeave"
        | "OnCancel"
        | "OnValidated";
    export interface ITransducerEvent<T extends THooks>
        extends IHookPayload<T> {
        readonly transducer: Transducer;
        readonly context: Context;
    }
    export interface ITransducerEnterEvent
        extends ITransducerEvent<"OnEnter"> {}
    export interface ITransducerSkipEvent extends ITransducerEvent<"OnSkip"> {}
    export interface ITransducerLeaveEvent
        extends ITransducerEvent<"OnLeave"> {}
    export interface ITransducerCancelEvent
        extends ITransducerEvent<"OnCancel"> {}
    export interface ITransducerValidatedEvent
        extends ITransducerEvent<"OnValidated"> {
        readonly result: "fail" | "pass";
        readonly type: "initial" | "revalidate";
    }
}

declare module "@tripetto/vfsm/lib/instance/data/events" {
    /** Dependencies */
    import { Data } from "@tripetto/vfsm/lib/instance/data/data";
    import { IHookPayload } from "@marblecore/hookup";
    import { Slot } from "@tripetto/slots";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    import { TSerializeTypes } from "@marblecore/serializer";
    import { ImmutableValue } from "@tripetto/vfsm/lib/instance/data/immutable";
    export type TValuesEvents = "OnValueChange";
    export interface IValueChangeEvent<
        T extends TSerializeTypes = TSerializeTypes,
        S extends Slot<T> = Slot<T>
    > extends IHookPayload<"OnValueChange"> {
        /** Reference to the transducer. */
        readonly transducer: Transducer;
        /** Reference to the slot. */
        readonly slot: S;
        /** Reference to the data instance that holds the all values. */
        readonly data: Data;
        /** Identifier of the value. */
        readonly id: number;
        /** Reference to the value. */
        readonly value: ImmutableValue<T, S>;
    }
}

declare module "@tripetto/vfsm/lib/factory" {
    /** Dependencies */
    import { StrictItem, StrictMatrix } from "@marblecore/matrix";
    import { THooks } from "@marblecore/hookup";
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { Branch } from "@tripetto/vfsm/lib/branches/branch";
    import { IOntologyProperties } from "@tripetto/vfsm/lib/interface";
    export interface IFactory<
        Dispatcher extends StrictMatrix<
            Dispatcher,
            Factory<Dispatcher, Ontology, Hooks>,
            void,
            Ontology
        >,
        T extends Factory<Dispatcher, Ontology, Hooks>,
        Ontology extends IOntologyProperties,
        Hooks extends THooks = void
    > {
        readonly property: string;
        new (pDispatcher: Dispatcher, pOntology: Ontology): T;
    }
    export abstract class Factory<
        Dispatcher extends StrictMatrix<
            Dispatcher,
            Factory<Dispatcher, Ontology, Hooks>,
            void,
            Ontology
        >,
        Ontology extends IOntologyProperties,
        Hooks extends THooks = void
    > extends StrictItem<Dispatcher, Hooks, Ontology> {
        /** Contains the ontology property. */
        static readonly property: string;
        /** Reference to the ancestor state. */
        readonly ancestor: State | undefined;
        /** Reference to the parent branch. */
        readonly branch: Branch | undefined;
        /**
         * Constructs a new item.
         * @param pDispatcher Reference to the dispatcher.
         * @param sHash Specifies the unique hash for the item.
         */
        constructor(pDispatcher: Dispatcher, sHash: string);
        /** Retrieves the identifier. */
        readonly id: string;
        /**
         * Sets the successor state.
         * @param sMode Specifies the mode.
         * @param pState Specifies the state to set.
         */
        setSuccessor(
            sMode: "live" | "preview",
            pState: State | undefined
        ): void;
        /**
         * Retrieves a reference to the successor state.
         * @param sMode Specifies the mode.
         * @return Returns the successor state.
         */
        successor(sMode: "live" | "preview"): State | undefined;
    }
}

declare module "@tripetto/vfsm/lib/branches/culling" {
    export type TBranchCulling = "first" | "each";
}

declare module "@tripetto/vfsm/lib/states/main" {
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { States } from "@tripetto/vfsm/lib/states/states";
    export class Main extends State {
        /**
         * Constructs a new main state.
         * @param pStates Reference to the state dispatcher.
         */
        constructor(pStates: States);
        /** Enter condition for the main state. */
        onEnterCondition(): boolean;
        /** Main state is entered. */
        onEnter(): void;
        /** Leave condition for the main state. */
        onLeaveCondition(): boolean;
        /** Main state is validated. */
        onValidate(): boolean;
    }
}

declare module "@tripetto/vfsm/lib/instance/moments/snapshot" {
    /** Dependencies */
    import { IStackSnapshot } from "@tripetto/vfsm/lib/instance/stack/snapshot";
    export interface IMomentSnapshot {
        /** Contains the stack snapshot. */
        readonly a: IStackSnapshot[];
        /** Contains the stack pointer. */
        readonly b: string;
    }
}

declare module "@tripetto/vfsm/lib/instance/orchestrator" {
    /** Dependencies */
    import { Condition } from "@tripetto/vfsm/lib/conditions/condition";
    import { IChainLookup } from "@tripetto/vfsm/lib/chain";
    import { Instance } from "@tripetto/vfsm/lib/instance/instance";
    import { States } from "@tripetto/vfsm/lib/states/states";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    export interface IInstanceOrchestrator {
        /** Contains the unique ontology fingerprint. */
        readonly fingerprint: string;
        /** Reference to the ontology states. */
        readonly states: States;
        /** Specifies if the automatic push-forward is enabled. */
        readonly autoPushForward: boolean;
        /** Specifies if the preview mode is enabled. */
        readonly preview: boolean;
        /** Chain. */
        readonly chain: IChainLookup;
        /** Contains the events. */
        readonly events: {
            /** Invoked when an instance is created. */
            onInstanceCreate: (pInstance: Instance) => void;
            /** Invoked when an instance is started. */
            onInstanceStart: (pInstance: Instance) => void;
            /** Invoked when an instance has finished. */
            onInstanceFinish: (pInstance: Instance) => void;
            /** Invoked when an instance is paused. */
            onInstancePause: (pInstance: Instance) => void;
            /** Invoked when an instance is stopped. */
            onInstanceStop: (pInstance: Instance) => void;
            /** Invoked when an instance ends. */
            onInstanceEnd: (
                pInstance: Instance,
                sType: "finished" | "stopped" | "paused"
            ) => void;
        };
        /** Spawns a new instance. */
        readonly spawn: (pInstance: Instance, nIndex: number) => Instance;
        /** Iterates through all possible transducers and their context. */
        readonly transducers: <T extends Transducer, C extends Condition>(
            fnTransducer: (pTransducer: T, pContext: C[]) => void
        ) => void;
        /** Logger. */
        readonly logger?: (sMessage: string) => void;
    }
}

declare module "@tripetto/vfsm/lib/threading" {
    export type TThreading = "single" | "multi";
}

declare module "@tripetto/vfsm/lib/instance/data/snapshot" {
    /** Dependencies */
    import { TSerializeTypes } from "@marblecore/serializer";
    export interface IValueSnapshot {
        /** Contains the context. */
        readonly a: string;
        /** Contains the value. */
        readonly b?: TSerializeTypes;
        /** Contains the reference. */
        readonly c?: string;
        /** Specifies value flags. */
        readonly d?: number;
        /** Contains the UTC set time of the value. */
        readonly e?: number;
    }
    export interface IDataSnapshot {
        /** Contains the store. */
        readonly a: string;
        /** Contains the values. */
        readonly b?: IValueSnapshot[];
    }
}

declare module "@tripetto/vfsm/lib/instance/stack/stack" {
    import { Branch } from "@tripetto/vfsm/lib/branches/branch";
    import { Condition } from "@tripetto/vfsm/lib/conditions/condition";
    import { Instance } from "@tripetto/vfsm/lib/instance/instance";
    import { IStackOrchestrator } from "@tripetto/vfsm/lib/instance/stack/orchestrator";
    import { IStackSnapshot } from "@tripetto/vfsm/lib/instance/stack/snapshot";
    import { Item, Matrix } from "@marblecore/matrix";
    import { Opcode } from "@tripetto/vfsm/enum/opcodes";
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    import { Validators } from "@tripetto/vfsm/lib/instance/validators/validators";
    export class Stack extends Item {
        /**
         * Creates a new stack.
         * @param pStack Reference to the stack matrix.
         * @param pOrchestrator Reference to the stack orchestrator.
         * @param pAncestor Reference to the stack ancestor.
         * @param pState Reference to a state.
         * @param pBranch Reference to a branch.
         * @param pCondition Reference to a condition.
         * @param pBranched Contains a reference to the active branch.
         * @param pValidators Reference to the transducer validators.
         * @param nOpcode Specifies the opcode.
         * @param nCounter Specifies the number of loops.
         * @param nSteps Specifies the number of steps.
         * @param bReturn Specifies if there is a return value.
         * @param bReturnOverride Specifies if a return override value is present.
         * @param pReturn Specifies the return state.
         * @param bBranching Specifies if the stack is branching.
         */
        static create(
            pStack: Matrix<Stack>,
            pOrchestrator: IStackOrchestrator,
            pState: State | undefined,
            pAncestor?: Stack | undefined,
            pBranch?: Branch | undefined,
            pCondition?: Condition | undefined,
            pBranched?: Branch | undefined,
            pValidators?: Validators | undefined,
            nOpcode?: Opcode,
            nCounter?: number,
            bReturn?: boolean,
            bReturnOverride?: boolean,
            pReturn?: State | undefined,
            bBranching?: boolean
        ): Stack;
        /** Retrieves the identifier. */
        readonly id: string;
        /** Retrieves the instance of the stack. */
        readonly instance: Instance;
        /** Contains a reference to the ancestor stack. */
        readonly ancestor: Stack | undefined;
        /** Reference to the state for the stack. */
        readonly state: State | undefined;
        /** Retrieves the validated transducers for the stack. */
        readonly transducers: Transducer[];
        /** Reference to the branch for the stack. */
        readonly branch: Branch | undefined;
        /** Reference to the condition for the stack. */
        readonly condition: Condition | undefined;
        /** Retrieves the mutable stack context. */
        readonly mutableContext: Condition | undefined;
        /** Retrieves the immutable stack context. */
        readonly immutableContext: Condition | undefined;
        /** Retrieves the stack checksum. */
        readonly checksum: string;
        /** Retrieves if the stack is branching. */
        readonly isBranch: boolean;
        /** Specifies if the stack is scheduled for the pursue step. */
        readonly isScheduled: boolean;
        /** Retrieves if the stack branch is at the finish. */
        readonly isAtFinish: boolean;
        /** Retrieves if the stack is evaluating. */
        readonly isEvaluating: boolean;
        /**
         * Performs a step.
         * @param nOpcode Specifies the opcode to be executed.
         */
        step(nOpcode?: Opcode): void;
        /**
         * Invalidates a transducer.
         * @param pTransducer Specifies the transducer to invalidate.
         */
        invalidate(pTransducer: Transducer): void;
        /** Cancels the stack. */
        cancel(): void;
        /**
         * Destroys the stack.
         * @return Returns `true` if the function succeeded.
         */
        destroy(): void;
        /**
         * Snapshots the stack.
         * @return Returns the stack snapshot.
         */
        snapshot(): IStackSnapshot;
    }
}

declare module "@tripetto/vfsm/lib/constructors" {
    /** Dependencies */
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { TState } from "@tripetto/vfsm/lib/states/type";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    import { TTransducer } from "@tripetto/vfsm/lib/transducers/type";
    import { Branch } from "@tripetto/vfsm/lib/branches/branch";
    import { TBranch } from "@tripetto/vfsm/lib/branches/type";
    import { Condition } from "@tripetto/vfsm/lib/conditions/condition";
    import { TCondition } from "@tripetto/vfsm/lib/conditions/type";
    export interface IConstructors<
        S extends State,
        T extends Transducer,
        B extends Branch,
        C extends Condition
    > {
        readonly state: TState<S>;
        readonly transducer: TTransducer<T>;
        readonly branch: TBranch<B>;
        readonly condition: TCondition<C>;
    }
}

declare module "@tripetto/vfsm/lib/branches/type" {
    /** Dependencies */
    import { Branch } from "@tripetto/vfsm/lib/branches/branch";
    import { Branches } from "@tripetto/vfsm/lib/branches/branches";
    import { IBranch } from "@tripetto/vfsm/lib/branches/interface";
    import { IFactory } from "@tripetto/vfsm/lib/factory";
    import { TAny } from "@marblecore/functional";
    import { TBranchEvents } from "@tripetto/vfsm/lib/branches/events";
    export type TBranch<
        T extends Branch = Branch,
        Ontology extends IBranch = TAny
    > = IFactory<Branches, T, Ontology, TBranchEvents>;
}

declare module "@tripetto/vfsm/lib/conditions/type" {
    /** Dependencies */
    import { Condition } from "@tripetto/vfsm/lib/conditions/condition";
    import { Conditions } from "@tripetto/vfsm/lib/conditions/conditions";
    import { ICondition } from "@tripetto/vfsm/lib/conditions/interface";
    import { IFactory } from "@tripetto/vfsm/lib/factory";
    import { TAny } from "@marblecore/functional";
    export type TCondition<
        T extends Condition = Condition,
        Ontology extends ICondition = TAny
    > = IFactory<Conditions, T, Ontology>;
}

declare module "@tripetto/vfsm/lib/states/type" {
    /** Dependencies */
    import { IFactory } from "@tripetto/vfsm/lib/factory";
    import { IState } from "@tripetto/vfsm/lib/states/interface";
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { States } from "@tripetto/vfsm/lib/states/states";
    import { TAny } from "@marblecore/functional";
    import { TStateEvents } from "@tripetto/vfsm/lib/states/events";
    export type TState<
        T extends State = State,
        Ontology extends IState = TAny
    > = IFactory<States, T, Ontology, TStateEvents>;
}

declare module "@tripetto/vfsm/lib/transducers/type" {
    /** Dependencies */
    import { IFactory } from "@tripetto/vfsm/lib/factory";
    import { ITransducer } from "@tripetto/vfsm/lib/transducers/interface";
    import { TAny } from "@marblecore/functional";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    import { Transducers } from "@tripetto/vfsm/lib/transducers/transducers";
    import { TTransducerEvents } from "@tripetto/vfsm/lib/transducers/events";
    export type TTransducer<
        T extends Transducer = Transducer,
        Ontology extends ITransducer = TAny
    > = IFactory<Transducers, T, Ontology, TTransducerEvents>;
}

declare module "@tripetto/vfsm/lib/instance/stack/snapshot" {
    /** Dependencies */
    import { IValidatorSnapshot } from "@tripetto/vfsm/lib/instance/validators/snapshot";
    import { Opcode } from "@tripetto/vfsm/enum/opcodes";
    export interface IStackSnapshot {
        /** Contains the stack. */
        readonly a: string;
        /** Contains the counter. */
        readonly b?: number;
        /** Contains the opcode. */
        readonly c?: Opcode;
        /** Contains the flags. */
        readonly d?: number;
        /** Contains the ancestor. */
        readonly e?: string;
        /** Contains the state. */
        readonly f?: string;
        /** Contains the branch. */
        readonly g?: string;
        /** Contains the active branch. */
        readonly h?: string;
        /** Contains the condition. */
        readonly i?: string;
        /** Contains the return state. */
        readonly j?: string;
        /** Contains the transducer snapshot. */
        readonly k?: IValidatorSnapshot[];
    }
}

declare module "@tripetto/vfsm/lib/instance/stack/orchestrator" {
    /** Dependencies */
    import { Branch } from "@tripetto/vfsm/lib/branches/branch";
    import { Condition } from "@tripetto/vfsm/lib/conditions/condition";
    import { IChainLookup } from "@tripetto/vfsm/lib/chain";
    import { Instance } from "@tripetto/vfsm/lib/instance/instance";
    import { Stack } from "@tripetto/vfsm/lib/instance/stack/stack";
    import { State } from "@tripetto/vfsm/lib/states/state";
    export interface IStackOrchestrator {
        /** Reference to the instance. */
        readonly instance: Instance;
        /** Contains if the preview mode is enabled. */
        readonly preview: boolean;
        /** Spawns a new instance. */
        readonly spawn: (pState: State) => void;
        /** Sets the validation state of the instance. */
        readonly validated: (sResult: "fail" | "pass") => void;
        /** Invoked when a stack is canceled. */
        readonly cancel: () => void;
        /** Snapshots. */
        readonly snapshot: {
            /** Pushes the current moment snapshot. */
            readonly push: () => void;
            /** Makes a snapshot of the current stack. */
            readonly make: () => void;
        };
        /** Pointer. */
        readonly pointer: {
            /** Retrieves the stack pointer. */
            readonly get: () => Stack | undefined;
            /** Assures a stack is bind to the pointer. */
            readonly assure: (pStack: Stack) => Stack | undefined;
            /** Releases the current pointer. */
            readonly release: () => void;
        };
        /** Chain. */
        readonly chain: IChainLookup;
        /** Logger. */
        readonly logger: (
            pStack: Stack,
            pAncestor: Stack | undefined,
            pState: State | undefined,
            pBranch: Branch | undefined,
            pCondition: Condition | undefined,
            pReturn: State | true | undefined
        ) => void;
    }
}

declare module "@tripetto/vfsm/enum/opcodes" {
    export const enum Opcode {
        Push = 0,
        Branch = 1,
        /**
         * An item is awaiting further processing after the branch conditions
         * are processed.
         */
        BranchWait = 2,
        Enter = 3,
        EnterCondition = 4,
        EnterConditionWait = 5,
        Wait = 6,
        LeaveCondition = 7,
        LeaveConditionWait = 8,
        Pursue = 9,
        Pop = 10
    }
}

declare module "@tripetto/vfsm/lib/instance/validators/validators" {
    /** Dependencies */
    import { Context } from "@tripetto/vfsm/lib/instance/data/context";
    import { IValidatorSnapshot } from "@tripetto/vfsm/lib/instance/validators/snapshot";
    import { Matrix } from "@marblecore/matrix";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    import { Validator } from "@tripetto/vfsm/lib/instance/validators/validator";
    export class Validators extends Matrix<Validator> {
        /** Retrieves the validated transducers. */
        readonly transducers: Transducer[];
        /** Retrieves if one or more validators are enabled. */
        readonly isEnabled: boolean;
        /** Retrieves if all validators passed the validation. */
        readonly isPassed: boolean;
        /**
         * Creates a new transducer validator.
         * @param pTransducer Reference to the transducer.
         * @param pContext Context for the validator.
         * @param bEnabled Specifies if the transducer is enabled.
         */
        create(
            pTransducer: Transducer,
            pContext: Context,
            bEnabled: boolean
        ): Validator;
        /**
         * Resumes a validator.
         * @param pTransducer Reference to the transducer.
         * @param pContext Context for the validator.
         * @param pSnapshot Reference to the snapshot data.
         */
        resume(
            pTransducer: Transducer,
            pContext: Context,
            pSnapshot: IValidatorSnapshot
        ): Validator;
        /**
         * Retrieves the validator for the specified transducer.
         * @param pTransducer Specifies the transducer.
         * @return Returns the validator instance or `undefined`.
         */
        forTransducer(pTransducer: Transducer): Validator | undefined;
    }
}

declare module "@tripetto/vfsm/lib/instance/validators/snapshot" {
    export interface IValidatorSnapshot {
        /** Contains the transducer. */
        readonly a: string;
        /** Contains if the validator flags. */
        readonly b: number;
    }
}

declare module "@tripetto/vfsm/lib/instance/validators/validator" {
    /** Dependencies */
    import { ICallbackPromise } from "@marblecore/callback";
    import { Context } from "@tripetto/vfsm/lib/instance/data/context";
    import { Item } from "@marblecore/matrix";
    import { IValidatorSnapshot } from "@tripetto/vfsm/lib/instance/validators/snapshot";
    import { State } from "@tripetto/vfsm/lib/states/state";
    import { Transducer } from "@tripetto/vfsm/lib/transducers/transducer";
    export class Validator extends Item {
        /**
         * Constructs a new transducer validator.
         * @param pTransducer Reference to the transducer.
         * @param pContext Context of the validator.
         * @param bEnabled Specifies if the transducer is enabled.
         * @param bPass Specifies if the validation is passed.
         */
        constructor(
            pTransducer: Transducer,
            pContext: Context,
            bEnabled: boolean,
            bPass?: boolean
        );
        /** Retrieves a reference to the transducer. */
        readonly transducer: Transducer;
        /** Specifies if the transducer is enabled. */
        readonly isEnabled: boolean;
        /** Retrieves if the transducer validation is passed. */
        readonly isPassed: boolean;
        /**
         * Validates the transducer.
         * @param pState Reference to the state.
         * @param sType Specifies the type of validation.
         * @param bPreview Specifies if the preview mode is enabled.
         * @param fnAssert Specifies the stack assert function.
         * @param fnDone Specifies the callback function that is invoked when the
         * validation is completed.
         * @return Returns `true` if the transducer was validated.
         */
        validate(
            pState: State | undefined,
            sType: "initial" | "revalidate",
            bPreview: boolean,
            fnAssert: (pPromise: ICallbackPromise<boolean>) => boolean,
            fnDone: () => void
        ): boolean;
        /** Invalidates the validation state. */
        invalidate(): void;
        /** Cancels a pending validation. */
        cancel(): void;
        /**
         * Snapshots the stack transducer validator.
         * @return Returns the stack transducer snapshot.
         */
        snapshot(): IValidatorSnapshot | undefined;
    }
}
